use env_logger;
use futures::stream::StreamExt;
use log::{debug, error, info, log_enabled, Level};
use netlink_packet_core::NetlinkPayload;
use netlink_packet_route::RouteNetlinkMessage as RtnlMessage;
use netlink_sys::{AsyncSocket, SocketAddr};
use rtnetlink::new_connection;

use dynip_cloudflare::{get_config_path, read_config, Config};

const RTNLGRP_LINK: u32 = 1;
const RTNLGRP_IPV4_IFADDR: u32 = 5;

const fn nl_mgrp(group: u32) -> u32 {
    if group > 31 {
        panic!("use netlink_sys::Socket::add_membership() for this group");
    }
    if group == 0 {
        0
    } else {
        1 << (group - 1)
    }
}

#[tokio::main]
async fn main() {
    env_logger::init();
    let config_path = match get_config_path().await {
        Ok(cp) => cp,
        Err(tried_paths) => {
            let extra: String = if !tried_paths.is_empty() {
                let joined = tried_paths
                    .iter()
                    .filter_map(|path| path.to_str())
                    .collect::<Vec<_>>()
                    .join(", ");
                format!(", tried the paths: '{}'", &joined)
            } else {
                String::with_capacity(0)
            };
            error!("Failed to find any config file{}", &extra);
            return;
        }
    };
    let config = match read_config(&config_path).await {
        Ok(config) => config,
        Err(e) => {
            error!("Failed to read and parse config file {:?}", &e);
            return;
        }
    };

    let (mut conn, mut _handle, mut messages) = new_connection().unwrap();
    let groups = nl_mgrp(RTNLGRP_LINK) | nl_mgrp(RTNLGRP_IPV4_IFADDR);

    let addr = SocketAddr::new(0, groups);

    if let Err(e) = conn.socket_mut().socket_mut().bind(&addr) {
        error!("Failed to bind to socket: {:?}", &e);
        return;
    }

    tokio::spawn(conn);
    info!("Listening for IPv4 address changes and interface connect/disconnect events...");

    while let Some((message, _)) = messages.next().await {
        match message.payload {
            NetlinkPayload::InnerMessage(RtnlMessage::NewAddress(msg)) => {
                if log_enabled!(Level::Debug) {
                    debug!("New IPv4 address message: {:?}", msg);
                } else {
                    info!("New IPv4 address");
                }
            }
            NetlinkPayload::InnerMessage(RtnlMessage::DelAddress(msg)) => {
                if log_enabled!(Level::Debug) {
                    debug!("Deleted IPv4 address message: {:?}", msg);
                } else {
                    info!("Deleted IPv4 address");
                }
            }
            NetlinkPayload::InnerMessage(RtnlMessage::NewLink(link)) => {
                if log_enabled!(Level::Debug) {
                    debug!("New link message (interface connected): {:?}", link);
                } else {
                    info!("New link (interface connected)");
                }
            }
            NetlinkPayload::InnerMessage(RtnlMessage::DelLink(link)) => {
                if log_enabled!(Level::Debug) {
                    debug!("Deleted link message (interface disconnected): {:?}", link);
                } else {
                    info!("Deleted link (interface disconnected)");
                }
            }
            _ => {
                if log_enabled!(Level::Debug) {
                    debug!("Unhandled message payload: {:?}", message.payload);
                }
            }
        }
    }
}
